import copy
import uuid
import uvicorn
from fastapi import FastAPI
from logging import getLogger
from typing import Dict, List, Optional

from llama_index.core import StorageContext, VectorStoreIndex
from llama_index.core.objects import ObjectIndex, SimpleObjectNodeMapping
from llama_index.core.storage.kvstore.types import BaseKVStore
from llama_index.core.storage.kvstore import SimpleKVStore
from llama_index.core.vector_stores.types import BasePydanticVectorStore

from llama_agents.control_plane.base import BaseControlPlane
from llama_agents.message_consumers.base import (
    BaseMessageQueueConsumer,
    StartConsumingCallable,
)
from llama_agents.message_consumers.callable import CallableMessageConsumer
from llama_agents.message_consumers.remote import RemoteMessageConsumer
from llama_agents.message_queues.base import BaseMessageQueue, PublishCallback
from llama_agents.messages.base import QueueMessage
from llama_agents.orchestrators.base import BaseOrchestrator
from llama_agents.tools import ServiceTool
from llama_agents.types import (
    ActionTypes,
    ServiceDefinition,
    TaskDefinition,
    TaskResult,
)

logger = getLogger(__name__)


class ControlPlaneServer(BaseControlPlane):
    """Control plane server.

    The control plane is responsible for managing the state of the system, including:
    - Registering services.
    - Submitting tasks.
    - Managing task state.
    - Handling service completion.
    - Launching the control plane server.

    Args:
        message_queue (BaseMessageQueue): Message queue for the system.
        orchestrator (BaseOrchestrator): Orchestrator for the system.
        vector_store (Optional[BasePydanticVectorStore], optional): Vector store for the system. Defaults to None.
        publish_callback (Optional[PublishCallback], optional): Callback for publishing messages. Defaults to None.
        state_store (Optional[BaseKVStore], optional): State store for the system. Defaults to None.
        services_store_key (str, optional): Key for the services store. Defaults to "services".
        tasks_store_key (str, optional): Key for the tasks store. Defaults to "tasks".
        step_interval (float, optional): The interval in seconds to poll for tool call results. Defaults to 0.1s.
        services_retrieval_threshold (int, optional): The threshold for retrieving services. Defaults to 5.
        host (str, optional): The host of the service. Defaults to "127.0.0.1".
        port (Optional[int], optional): The port of the service. Defaults to 8000.
        running (bool, optional): Whether the service is running. Defaults to True.

    Examples:
        ```python
        from llama_agents import ControlPlaneServer
        from llama_agents import SimpleMessageQueue, AgentOrchestrator
        from llama_index.llms.openai import OpenAI

        control_plane = ControlPlaneServer(
            SimpleMessageQueue(),
            AgentOrchestrator(llm=OpenAI())
        )
        ```
    """

    def __init__(
        self,
        message_queue: BaseMessageQueue,
        orchestrator: BaseOrchestrator,
        vector_store: Optional[BasePydanticVectorStore] = None,
        publish_callback: Optional[PublishCallback] = None,
        state_store: Optional[BaseKVStore] = None,
        services_store_key: str = "services",
        tasks_store_key: str = "tasks",
        step_interval: float = 0.1,
        services_retrieval_threshold: int = 5,
        host: str = "127.0.0.1",
        port: Optional[int] = 8000,
        running: bool = True,
    ) -> None:
        self.orchestrator = orchestrator
        self.object_index = ObjectIndex(
            VectorStoreIndex(
                nodes=[],
                storage_context=StorageContext.from_defaults(vector_store=vector_store),
            ),
            SimpleObjectNodeMapping(),
        )
        self.step_interval = step_interval
        self.running = running
        self.host = host
        self.port = port

        self.state_store = state_store or SimpleKVStore()

        # TODO: should we store services in a tool retriever?
        self.services_store_key = services_store_key
        self.tasks_store_key = tasks_store_key

        self._message_queue = message_queue
        self._publisher_id = f"{self.__class__.__qualname__}-{uuid.uuid4()}"
        self._publish_callback = publish_callback

        self._services_cache: Dict[str, ServiceDefinition] = {}
        self._total_services = 0
        self._services_retrieval_threshold = services_retrieval_threshold

        self.app = FastAPI()
        self.app.add_api_route("/", self.home, methods=["GET"], tags=["Control Plane"])
        self.app.add_api_route(
            "/process_message",
            self.process_message,
            methods=["POST"],
            tags=["Control Plane"],
        )

        self.app.add_api_route(
            "/services/register",
            self.register_service,
            methods=["POST"],
            tags=["Services"],
        )
        self.app.add_api_route(
            "/services/deregister",
            self.deregister_service,
            methods=["POST"],
            tags=["Services"],
        )
        self.app.add_api_route(
            "/services/{service_name}",
            self.get_service,
            methods=["GET"],
            tags=["Services"],
        )
        self.app.add_api_route(
            "/services",
            self.get_all_services,
            methods=["GET"],
            tags=["Services"],
        )

        self.app.add_api_route(
            "/tasks", self.create_task, methods=["POST"], tags=["Tasks"]
        )
        self.app.add_api_route(
            "/tasks", self.get_all_tasks, methods=["GET"], tags=["Tasks"]
        )
        self.app.add_api_route(
            "/tasks/{task_id}",
            self.get_task_state_api_safe,
            methods=["GET"],
            tags=["Tasks"],
        )

    @property
    def message_queue(self) -> BaseMessageQueue:
        return self._message_queue

    @property
    def publisher_id(self) -> str:
        return self._publisher_id

    @property
    def publish_callback(self) -> Optional[PublishCallback]:
        return self._publish_callback

    async def process_message(self, message: QueueMessage) -> None:
        action = message.action

        if action == ActionTypes.NEW_TASK and message.data is not None:
            await self.create_task(TaskDefinition(**message.data))
        elif action == ActionTypes.COMPLETED_TASK and message.data is not None:
            await self.handle_service_completion(TaskResult(**message.data))
        else:
            raise ValueError(f"Action {action} not supported by control plane")

    def as_consumer(self, remote: bool = False) -> BaseMessageQueueConsumer:
        if remote:
            return RemoteMessageConsumer(
                id_=self.publisher_id,
                url=(
                    f"http://{self.host}:{self.port}/process_message"
                    if self.port
                    else f"http://{self.host}/process_message"
                ),
                message_type="control_plane",
            )

        return CallableMessageConsumer(
            id_=self.publisher_id,
            message_type="control_plane",
            handler=self.process_message,
        )

    async def launch_server(self) -> None:
        logger.info(f"Launching control plane server at {self.host}:{self.port}")
        # uvicorn.run(self.app, host=self.host, port=self.port)

        class CustomServer(uvicorn.Server):
            def install_signal_handlers(self) -> None:
                pass

        cfg = uvicorn.Config(self.app, host=self.host, port=self.port)
        server = CustomServer(cfg)
        await server.serve()

    async def home(self) -> Dict[str, str]:
        return {
            "running": str(self.running),
            "step_interval": str(self.step_interval),
            "services_store_key": self.services_store_key,
            "total_services": str(self._total_services),
            "services_retrieval_threshold": str(self._services_retrieval_threshold),
        }

    async def register_service(self, service_def: ServiceDefinition) -> None:
        await self.state_store.aput(
            service_def.service_name,
            service_def.model_dump(),
            collection=self.services_store_key,
        )

        # decide to use cache vs. retrieval
        self._total_services += 1
        if self._total_services > self._services_retrieval_threshold:
            # TODO: currently blocking, should be async
            self.object_index.insert_object(service_def.model_dump())
            for service in self._services_cache.values():
                self.object_index.insert_object(service.model_dump())
            self._services_cache = {}
        else:
            self._services_cache[service_def.service_name] = service_def

    async def deregister_service(self, service_name: str) -> None:
        deleted = await self.state_store.adelete(
            service_name, collection=self.services_store_key
        )
        if service_name in self._services_cache:
            del self._services_cache[service_name]

        if deleted:
            self._total_services -= 1
        # TODO: object index does not have delete yet

    async def get_service(self, service_name: str) -> ServiceDefinition:
        service_dict = await self.state_store.aget(
            service_name, collection=self.services_store_key
        )
        if service_dict is None:
            raise ValueError(f"Service with name {service_name} not found")

        return ServiceDefinition.model_validate(service_dict)

    async def get_all_services(self) -> Dict[str, ServiceDefinition]:
        service_dicts = await self.state_store.aget_all(
            collection=self.services_store_key
        )
        return {
            service_name: ServiceDefinition.model_validate(service_dict)
            for service_name, service_dict in service_dicts.items()
        }

    async def create_task(self, task_def: TaskDefinition) -> Dict[str, str]:
        await self.state_store.aput(
            task_def.task_id, task_def.model_dump(), collection=self.tasks_store_key
        )

        task_def = await self.send_task_to_service(task_def)
        await self.state_store.aput(
            task_def.task_id, task_def.model_dump(), collection=self.tasks_store_key
        )
        logger.debug(f"Task {task_def.task_id} created")
        return {"task_id": task_def.task_id}

    async def send_task_to_service(self, task_def: TaskDefinition) -> TaskDefinition:
        if self._total_services > self._services_retrieval_threshold:
            service_retriever = self.object_index.as_retriever(similarity_top_k=5)

            # could also route based on similarity alone.
            # TODO: Figure out user-specified routing
            service_def_dicts: List[dict] = await service_retriever.aretrieve(
                task_def.input
            )
            service_defs = [
                ServiceDefinition.model_validate(service_def_dict)
                for service_def_dict in service_def_dicts
            ]
        else:
            service_defs = list(self._services_cache.values())

        service_tools = [
            ServiceTool.from_service_definition(service_def)
            for service_def in service_defs
        ]

        next_messages, task_state = await self.orchestrator.get_next_messages(
            task_def, service_tools, task_def.state
        )

        logger.debug(f"Sending task {task_def.task_id} to services: {next_messages}")

        for message in next_messages:
            await self.publish(message)

        task_def.state.update(task_state)
        return task_def

    async def handle_service_completion(
        self,
        task_result: TaskResult,
    ) -> None:
        # add result to task state
        task_def = await self.get_task_state(task_result.task_id)
        state = await self.orchestrator.add_result_to_state(task_result, task_def.state)
        task_def.state.update(state)

        # generate and send new tasks (if any)
        task_def = await self.send_task_to_service(task_def)

        await self.state_store.aput(
            task_def.task_id, task_def.model_dump(), collection=self.tasks_store_key
        )

    async def get_task_state(self, task_id: str) -> TaskDefinition:
        state_dict = await self.state_store.aget(
            task_id, collection=self.tasks_store_key
        )
        if state_dict is None:
            raise ValueError(f"Task with id {task_id} not found")

        return TaskDefinition(**state_dict)

    async def get_task_state_api_safe(self, task_id: str) -> TaskDefinition:
        state_dict = await self.state_store.aget(
            task_id, collection=self.tasks_store_key
        )
        state_dict = copy.deepcopy(state_dict)

        if state_dict is None:
            raise ValueError(f"Task with id {task_id} not found")

        # remove an bytes objects from state
        for key, val in state_dict["state"].items():
            if isinstance(val, bytes):
                state_dict["state"][key] = "<bytes object>"

        return TaskDefinition(**state_dict)

    async def get_all_tasks(self) -> Dict[str, TaskDefinition]:
        state_dicts = await self.state_store.aget_all(collection=self.tasks_store_key)
        state_dicts = copy.deepcopy(state_dicts)

        task_defs = {}
        for task_id, state_dict in state_dicts.items():
            # remove an bytes objects from state
            for key, val in state_dict["state"].items():
                if isinstance(val, bytes):
                    state_dict["state"][key] = "<bytes object>"
            task_defs[task_id] = TaskDefinition(**state_dict)

        return task_defs

    async def register_to_message_queue(self) -> StartConsumingCallable:
        return await self.message_queue.register_consumer(self.as_consumer(remote=True))


if __name__ == "__main__":
    from llama_agents import SimpleMessageQueue, AgentOrchestrator
    from llama_index.llms.openai import OpenAI

    control_plane = ControlPlaneServer(
        SimpleMessageQueue(), AgentOrchestrator(llm=OpenAI())
    )
    import asyncio

    asyncio.run(control_plane.launch_server())



from llama_agents.orchestrators.agent import AgentOrchestrator
from llama_agents.orchestrators.base import BaseOrchestrator
from llama_agents.orchestrators.pipeline import PipelineOrchestrator

__all__ = [
    "BaseOrchestrator",
    "PipelineOrchestrator",
    "AgentOrchestrator",
]



import httpx
from typing import Any, List

from textual.app import ComposeResult
from textual.containers import VerticalScroll
from textual.reactive import reactive
from textual.widgets import Button, Static

from llama_agents.app.components.types import ButtonType


class ServiceButton(Button):
    type: ButtonType = ButtonType.SERVICE


class ServicesList(Static):
    services: List[str] = reactive([])

    def __init__(self, control_plane_url: str, **kwargs: Any):
        self.control_plane_url = control_plane_url
        super().__init__(**kwargs)

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="services-scroll"):
            for service in self.services:
                yield ServiceButton(service)

    async def on_mount(self) -> None:
        self.set_interval(2, self.refresh_services)

    async def refresh_services(self) -> None:
        async with httpx.AsyncClient(timeout=120.0) as client:
            response = await client.get(f"{self.control_plane_url}/services")
            services_dict = response.json()

        new_services = []
        for service_name in services_dict:
            new_services.append(service_name)

        self.services = [*new_services]

    async def watch_services(self, new_services: List[str]) -> None:
        try:
            services_scroll = self.query_one("#services-scroll")
            await services_scroll.remove_children()
            for service in new_services:
                await services_scroll.mount(ServiceButton(service))
        except Exception:
            pass



from llama_agents.tools.agent_service_tool import AgentServiceTool
from llama_agents.tools.meta_service_tool import MetaServiceTool
from llama_agents.tools.service_tool import ServiceTool
from llama_agents.tools.service_component import ServiceComponent


__all__ = ["AgentServiceTool", "MetaServiceTool", "ServiceTool", "ServiceComponent"]



"""
What does the processing loop for the control plane look like?
- check message queue
- handle incoming new tasks
- handle incoming general chats
- handle services returning a completed task
"""

from abc import ABC, abstractmethod
from typing import Dict

from llama_agents.message_queues.base import BaseMessageQueue
from llama_agents.message_consumers.base import (
    BaseMessageQueueConsumer,
    StartConsumingCallable,
)
from llama_agents.message_publishers.publisher import MessageQueuePublisherMixin
from llama_agents.types import (
    ServiceDefinition,
    TaskDefinition,
    TaskResult,
)


class BaseControlPlane(MessageQueuePublisherMixin, ABC):
    """The control plane for the system.

    The control plane is responsible for managing the state of the system, including:
    - Registering services.
    - Submitting tasks.
    - Managing task state.
    - Handling service completion.
    - Launching the control plane server.
    """

    @property
    @abstractmethod
    def message_queue(self) -> BaseMessageQueue:
        """Return associated message queue."""

    @abstractmethod
    def as_consumer(self, remote: bool = False) -> BaseMessageQueueConsumer:
        """
        Get the consumer for the message queue.

        Args:
            remote (bool):
                Whether the consumer is remote.
                If True, the consumer will be a RemoteMessageConsumer.

        Returns:
            BaseMessageQueueConsumer: Message queue consumer.
        """
        ...

    @abstractmethod
    async def register_service(self, service_def: ServiceDefinition) -> None:
        """
        Register an service with the control plane.

        Args:
            service_def (ServiceDefinition): Definition of the service.
        """
        ...

    @abstractmethod
    async def deregister_service(self, service_name: str) -> None:
        """
        Deregister a service from the control plane.

        Args:
            service_name (str): Name of the service.
        """
        ...

    @abstractmethod
    async def create_task(self, task_def: TaskDefinition) -> Dict[str, str]:
        """
        Submit a task to the control plane.

        Args:
            task_def (TaskDefinition): Definition of the task.

        Returns:
            dict: Task ID.
        """
        ...

    @abstractmethod
    async def send_task_to_service(self, task_def: TaskDefinition) -> TaskDefinition:
        """
        Send a task to an service.

        Args:
            task_def (TaskDefinition): Definition of the task.

        Returns:
            TaskDefinition: Task definition with updated state.
        """
        ...

    @abstractmethod
    async def handle_service_completion(
        self,
        task_result: TaskResult,
    ) -> None:
        """
        Handle the completion of a task by an service.

        Args:
            task_result (TaskResult): Result of the task.
        """
        ...

    @abstractmethod
    async def get_task_state(self, task_id: str) -> dict:
        """
        Get the current state of a task.

        Args:
            task_id (str): Unique identifier of the task.

        Returns:
            dict: State of the task
        """
        ...

    @abstractmethod
    async def get_all_tasks(self) -> dict:
        """
        Get all tasks.

        Returns:
            dict: All tasks.
        """
        ...

    @abstractmethod
    async def launch_server(self) -> None:
        """
        Launch the control plane server.
        """
        ...

    @abstractmethod
    async def register_to_message_queue(self) -> StartConsumingCallable:
        """Register the service to the message queue."""



import httpx
from typing import Dict, Optional, Union

from llama_agents.types import TaskDefinition, ServiceDefinition, TaskResult

DEFAULT_TIMEOUT = 120.0


class AsyncLlamaAgentsClient:
    """Client for interacting with the Llama Agents control plane asynchronously.

    Args:
        control_plane_url (str): The URL of the control plane.
        timeout (float, optional): The timeout for requests. Defaults to 120.0.
    """

    def __init__(self, control_plane_url: str, timeout: float = DEFAULT_TIMEOUT):
        self.control_plane_url = control_plane_url
        self.timeout = timeout

    async def create_task(self, task_def: Union[str, TaskDefinition]) -> str:
        """Create a new task with the control plane.

        Args:
            task_def (Union[str, TaskDefinition]):
                The task definition to create.
                If a string is provided, it will be used as the input for the task.

        Returns:
            str: The ID of the created task.
        """
        if isinstance(task_def, str):
            task_def = TaskDefinition(input=task_def)

        async with httpx.AsyncClient(timeout=self.timeout) as client:
            response = await client.post(
                f"{self.control_plane_url}/tasks", json=task_def.dict()
            )
            return response.json()["task_id"]

    async def get_tasks(self) -> Dict[str, TaskDefinition]:
        """Get all tasks registered with the control plane.

        Returns:
            Dict[str, TaskDefinition]: A dictionary of task IDs to task definitions
        """
        async with httpx.AsyncClient(timeout=self.timeout) as client:
            response = await client.get(f"{self.control_plane_url}/tasks")
            task_dicts = response.json()

        return {
            task_id: TaskDefinition(**task_dict)
            for task_id, task_dict in task_dicts.items()
        }

    async def get_task(self, task_id: str) -> TaskDefinition:
        """Get a task by ID.

        Args:
            task_id (str): The ID of the task to get.

        Returns:
            TaskDefinition: The definition of the task.
        """
        async with httpx.AsyncClient(timeout=self.timeout) as client:
            response = await client.get(f"{self.control_plane_url}/tasks/{task_id}")
            return TaskDefinition(**response.json())

    async def get_services(self) -> Dict[str, ServiceDefinition]:
        """Get all services registered with the control plane.

        Returns:
            Dict[str, ServiceDefinition]: A dictionary of service names to service definitions
        """
        async with httpx.AsyncClient(timeout=self.timeout) as client:
            response = await client.get(f"{self.control_plane_url}/services")
            service_def_dicts = response.json()

        return {
            service_name: ServiceDefinition(**service_def_dict)
            for service_name, service_def_dict in service_def_dicts.items()
        }

    async def get_service(self, service_name: str) -> ServiceDefinition:
        """Get the definition of a service by name.

        Args:
            service_name (str): The name of the service to get.

        Returns:
            ServiceDefinition: The definition of the service.
        """
        async with httpx.AsyncClient(timeout=self.timeout) as client:
            response = await client.get(
                f"{self.control_plane_url}/services/{service_name}"
            )
            return ServiceDefinition(**response.json())

    async def register_service(self, service_def: ServiceDefinition) -> None:
        """Register a service with the control plane.

        Args:
            service_def (ServiceDefinition): The service definition to register.

        Returns:
            None
        """
        async with httpx.AsyncClient(timeout=self.timeout) as client:
            await client.post(
                f"{self.control_plane_url}/services", json=service_def.model_dump()
            )

    async def deregister_service(self, service_name: str) -> None:
        """Deregister a service from the control plane.

        Args:
            service_name (str): The name of the service to deregister.

        Returns:
            None
        """
        async with httpx.AsyncClient(timeout=self.timeout) as client:
            await client.delete(f"{self.control_plane_url}/services/{service_name}")

    async def get_task_result(self, task_id: str) -> Optional[TaskResult]:
        """Get the result of a task if it has one.

        Args:
            task_id (str): The ID of the task to get the result for.

        Returns:
            Optional[TaskResult]: The result of the task if it has one.
        """
        task = await self.get_task(task_id)

        try:
            return TaskResult(**task.state["result"])
        except KeyError:
            raise ValueError(f"Task {task_id} does not have a result yet.")



import asyncio
import signal
import sys
import uuid
from typing import Any, Callable, Dict, List, Optional

from llama_agents.services.base import BaseService
from llama_agents.control_plane.base import BaseControlPlane
from llama_agents.message_consumers.base import (
    BaseMessageQueueConsumer,
    StartConsumingCallable,
)
from llama_agents.message_queues.simple import SimpleMessageQueue
from llama_agents.message_queues.base import BaseMessageQueue
from llama_agents.message_queues.base import PublishCallback
from llama_agents.messages.base import QueueMessage
from llama_agents.types import ActionTypes, TaskDefinition, TaskResult
from llama_agents.message_publishers.publisher import MessageQueuePublisherMixin


class HumanMessageConsumer(BaseMessageQueueConsumer):
    message_handler: Dict[str, Callable]
    message_type: str = "human"

    async def _process_message(self, message: QueueMessage, **kwargs: Any) -> None:
        action = message.action
        if action not in self.message_handler:
            raise ValueError(f"Action {action} not supported by control plane")

        if action == ActionTypes.COMPLETED_TASK:
            await self.message_handler[action](message_data=message.data)


class LocalLauncher(MessageQueuePublisherMixin):
    """Launches a llama-agents system locally, in a single async loop.

    The LocalLauncher is a convenience class for launching a system of services.

    When launching, the launcher will
    - Register each service to the control plane.
    - Start each service.
    - Register a human consumer to handle messages.
    - Publish an initial task to the control plane.
    - Run until the a result is found.
    - Clean up registered services by deregistering them from the control plane.
    - Clean up consumers by deregistering them from the message queue.

    Args:
        services (List[BaseService]):
            List of services to launch.
        control_plane (BaseControlPlane):
            Control plane for the system.
        message_queue (SimpleMessageQueue):
            Message queue for the system.
        publish_callback (Optional[PublishCallback], optional):
            Callback for publishing messages. Defaults to None.

    Examples:
        ```python
        from llama_agents import LocalLauncher

        launcher = LocalLauncher(
            services=[service1, service2],
            control_plane=control_plane,
            message_queue=message_queue,
        )

        # sync
        result = launcher.launch_single("Do the thing.")

        # async
        result = await launcher.alaunch_single("Do the thing again.")
        ```
    """

    def __init__(
        self,
        services: List[BaseService],
        control_plane: BaseControlPlane,
        message_queue: BaseMessageQueue = SimpleMessageQueue(),
        publish_callback: Optional[PublishCallback] = None,
    ) -> None:
        self.services = services
        self.control_plane = control_plane
        self._message_queue = message_queue
        self._publisher_id = f"{self.__class__.__qualname__}-{uuid.uuid4()}"
        self._publish_callback = publish_callback
        self.result: Optional[str] = None

    @property
    def message_queue(self) -> SimpleMessageQueue:
        """Message queue."""
        return self._message_queue

    @property
    def publisher_id(self) -> str:
        """Publisher ID."""
        return self._publisher_id

    @property
    def publish_callback(self) -> Optional[PublishCallback]:
        """Publish callback, if any."""
        return self._publish_callback

    async def handle_human_message(self, **kwargs: Any) -> None:
        result = TaskResult(**kwargs["message_data"])
        self.result = result.result

    async def register_consumers(
        self, consumers: Optional[List[BaseMessageQueueConsumer]] = None
    ) -> List[StartConsumingCallable]:
        start_consuming_callables = []
        for service in self.services:
            start_consuming_callables.append(
                await self.message_queue.register_consumer(service.as_consumer())
            )

        consumers = consumers or []
        for consumer in consumers:
            start_consuming_callables.append(
                await self.message_queue.register_consumer(consumer)
            )

        start_consuming_callables.append(
            await self.message_queue.register_consumer(self.control_plane.as_consumer())
        )
        return start_consuming_callables

    def launch_single(self, initial_task: str) -> str:
        """Launch the system in a single async loop."""
        return asyncio.run(self.alaunch_single(initial_task))

    def get_shutdown_handler(self, tasks: List[asyncio.Task]) -> Callable:
        def signal_handler(sig: Any, frame: Any) -> None:
            print("\nShutting down.")
            for task in tasks:
                task.cancel()
            sys.exit(0)

        return signal_handler

    async def alaunch_single(self, initial_task: str) -> str:
        """Launch the system in a single async loop."""
        # clear any result
        self.result = None

        # register human consumer
        human_consumer = HumanMessageConsumer(
            message_handler={
                ActionTypes.COMPLETED_TASK: self.handle_human_message,
            }
        )
        start_consuming_callables = await self.register_consumers([human_consumer])

        # register each service to the control plane
        for service in self.services:
            await self.control_plane.register_service(service.service_definition)

        # start services
        bg_tasks: List[asyncio.Task] = []
        for service in self.services:
            if hasattr(service, "raise_exceptions"):
                service.raise_exceptions = True  # ensure exceptions are raised
            bg_tasks.append(await service.launch_local())

        # consumers start consuming in their own threads
        start_consuming_tasks = []
        for start_consuming_callable in start_consuming_callables:
            task = asyncio.create_task(start_consuming_callable())
            start_consuming_tasks.append(task)

        # publish initial task
        await self.publish(
            QueueMessage(
                type="control_plane",
                action=ActionTypes.NEW_TASK,
                data=TaskDefinition(input=initial_task).model_dump(),
            ),
        )
        # runs until the message queue is stopped by the human consumer
        mq_task = await self.message_queue.launch_local()
        shutdown_handler = self.get_shutdown_handler(
            [mq_task] + bg_tasks + start_consuming_tasks
        )
        loop = asyncio.get_event_loop()
        while loop.is_running():
            await asyncio.sleep(0.1)
            signal.signal(signal.SIGINT, shutdown_handler)

            for task in bg_tasks:
                if task.done() and task.exception():  # type: ignore
                    raise task.exception()  # type: ignore

            if mq_task is not None and mq_task.done() and mq_task.exception():  # type: ignore
                raise mq_task.exception()  # type: ignore

            if self.result:
                break

        # shutdown tasks
        for task in bg_tasks + start_consuming_tasks:
            task.cancel()

        # clean up registered services
        for service in self.services:
            await self.control_plane.deregister_service(
                service.service_definition.service_name
            )

        # clean up consumers
        for service in self.services:
            await self.message_queue.deregister_consumer(service.as_consumer())

        await self.message_queue.deregister_consumer(human_consumer)
        await self.message_queue.deregister_consumer(self.control_plane.as_consumer())

        # clean up before shutting down mq
        await self.message_queue.cleanup_local(
            message_types=[
                c.message_type
                for c in [s.as_consumer() for s in self.services]
                + [self.control_plane.as_consumer(), human_consumer]
            ]
        )
        mq_task.cancel()

        return self.result or "No result found."



from typing import Any, Dict, List, Tuple

from llama_index.core.llms import LLM
from llama_index.core.memory import ChatMemoryBuffer
from llama_index.core.tools import BaseTool

from llama_agents.messages.base import QueueMessage
from llama_agents.orchestrators.base import BaseOrchestrator
from llama_agents.tools.service_tool import ServiceTool
from llama_agents.types import ActionTypes, ChatMessage, TaskDefinition, TaskResult

import logging

logger = logging.getLogger(__name__)

HISTORY_KEY = "chat_history"
RESULT_KEY = "result"
DEFAULT_SUMMARIZE_TMPL = "{history}\n\nThe above represents the progress so far, please condense the messages into a single message."
DEFAULT_FOLLOWUP_TMPL = (
    "Pick the next action to take. Invoke the 'finalize' tool with your full final answer if the answer to the original input is in the chat history. "
    "As a reminder, the original input was: {original_input}"
)


class AgentOrchestrator(BaseOrchestrator):
    """An orchestrator that uses an agentic LLM to process messages.

    Each incoming task is processed by the LLM, which decides which service to call next.

    If no tools are needed, or the LLm deems the task complete, the orchestrator will finalize
    the task by sending a message to the `human` message queue for final processing.

    Attributes:
        llm (LLM):
            The LLM to use for processing messages.
        human_description (str):
            A human-readable description of the finalize tool.
        summarize_prompt (str):
            The prompt to use for summarizing the chat history.
        followup_prompt (str):
            The prompt to use for prompting the user for the next action to take.

    Examples:
        ```python
        from llama_agents import AgentOrchestrator

        # orchestrator with default prompts
        orchestrator = AgentOrchestrator(
            llm=llm,
            human_description=(
                "Useful for finalizing a response.
                Should contain a complete answer that satisfies the original input."
            ),
            summarize_prompt=(
                "{history}\n\nThe above represents the progress so far, please condense "
                "the messages into a single message."
            ),
            followup_prompt=(
                "Pick the next action to take. Invoke the 'finalize' tool with your full final answer "
                "if the answer to the original input is in the chat history. "
                "As a reminder, the original input was: {original_input}"
            ),
        )
        ```
    """

    def __init__(
        self,
        llm: LLM,
        human_description: str = "Useful for finalizing a response. Should contain a complete answer that satisfies the original input.",
        summarize_prompt: str = DEFAULT_SUMMARIZE_TMPL,
        followup_prompt: str = DEFAULT_FOLLOWUP_TMPL,
    ):
        self.llm = llm
        self.summarize_prompt = summarize_prompt
        self.followup_prompt = followup_prompt
        self.finalize_tool = ServiceTool(name="finalize", description=human_description)

    async def get_next_messages(
        self, task_def: TaskDefinition, tools: List[BaseTool], state: Dict[str, Any]
    ) -> Tuple[List[QueueMessage], Dict[str, Any]]:
        """Get the next message to process. Returns the message and the new state.

        Uses the llm and the chat history related to the task to determine the next action to take.
        """
        tools_plus_human = [self.finalize_tool, *tools]

        chat_dicts = state.get(HISTORY_KEY, [])
        chat_history = [ChatMessage(**x) for x in chat_dicts]

        # TODO: how to make memory configurable?
        memory = ChatMemoryBuffer.from_defaults(chat_history=chat_history, llm=self.llm)

        # check if first message
        if len(chat_history) == 0:
            logger.debug("Agent input: " + task_def.input)
            memory.put(ChatMessage(role="user", content=task_def.input))
            response = await self.llm.apredict_and_call(
                tools,
                user_msg=task_def.input,
                error_on_no_tool_call=False,
            )
        else:
            messages = memory.get()
            logger.debug("Agent input: " + str(messages))
            response = await self.llm.apredict_and_call(
                tools_plus_human,
                chat_history=messages,
                error_on_no_tool_call=False,
            )

        # check if there was a tool call
        queue_messages = []
        result = None
        logger.debug(f"Agent response sources: {response.sources}")
        if len(response.sources) == 0 or response.sources[0].tool_name == "finalize":
            # convert memory chat messages
            llama_messages = memory.get_all()
            history = [ChatMessage(**x.dict()) for x in llama_messages]

            result = TaskResult(
                task_id=task_def.task_id, history=history, result=response.response
            )

            queue_messages.append(
                QueueMessage(
                    type="human",
                    data=result.model_dump(),
                    action=ActionTypes.COMPLETED_TASK,
                )
            )
        else:
            for source in response.sources:
                name = source.tool_name
                input_data = source.raw_input
                input_str = next(iter(input_data.values()))
                queue_messages.append(
                    QueueMessage(
                        type=name,
                        data=TaskDefinition(
                            task_id=task_def.task_id, input=input_str
                        ).model_dump(),
                        action=ActionTypes.NEW_TASK,
                    )
                )

        new_state = {
            HISTORY_KEY: [x.dict() for x in memory.get_all()],
            RESULT_KEY: result.model_dump() if result is not None else None,
        }
        return queue_messages, new_state

    async def add_result_to_state(
        self,
        result: TaskResult,
        state: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Add the result of processing a message to the state. Returns the new state."""

        # summarize the result
        new_history = result.history
        new_history_str = "\n".join([str(x) for x in new_history])
        # TODO: Better logic for when to summarize?
        if len(new_history) > 1:
            summarize_prompt_str = self.summarize_prompt.format(history=new_history_str)
            summary = await self.llm.acomplete(summarize_prompt_str)

        # get the current chat history, add the summary to it
        chat_dicts = state.get(HISTORY_KEY, [])
        chat_history = [ChatMessage(**x) for x in chat_dicts]

        chat_history.append(ChatMessage(role="assistant", content=str(summary)))

        # add the followup prompt to the chat history
        original_input = chat_history[0].content
        chat_history.append(
            ChatMessage(
                role="user",
                content=self.followup_prompt.format(original_input=original_input),
            )
        )

        new_state = {HISTORY_KEY: [x.dict() for x in chat_history]}
        return new_state



import httpx
from typing import Any, List

from textual.app import ComposeResult
from textual.containers import VerticalScroll, Container
from textual.reactive import reactive
from textual.widgets import Button, Static, Input

from llama_agents.app.components.types import ButtonType
from llama_agents.types import HumanResponse, TaskDefinition


class HumanTaskButton(Button):
    type: ButtonType = ButtonType.HUMAN
    task_id: str = ""


class HumanTaskList(Static):
    tasks: List[TaskDefinition] = reactive([])
    selected_task: str = reactive("")
    selected_task_id: str = reactive("")

    def __init__(self, human_service_url: str, **kwargs: Any):
        self.human_service_url = human_service_url
        super().__init__(**kwargs)

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="human-tasks-scroll"):
            for task in self.tasks:
                button = HumanTaskButton(task.input)
                button.task_id = task.task_id
                yield button

    async def on_mount(self) -> None:
        self.set_interval(2, self.refresh_tasks)

    async def refresh_tasks(self) -> None:
        async with httpx.AsyncClient(timeout=120.0) as client:
            response = await client.get(f"{self.human_service_url}/tasks")
            tasks = response.json()

        new_tasks = []
        for task in tasks:
            new_tasks.append(TaskDefinition(**task))

        self.tasks = [*new_tasks]

    async def watch_tasks(self, new_tasks: List[TaskDefinition]) -> None:
        try:
            tasks_scroll = self.query_one("#human-tasks-scroll")
            await tasks_scroll.remove_children()
            for task in new_tasks:
                button = HumanTaskButton(task.input)
                button.task_id = task.task_id
                await tasks_scroll.mount(button)
        except Exception:
            pass

    async def watch_selected_task(self, new_task: str) -> None:
        if not new_task:
            return

        try:
            await self.query_one("#respond").remove()
        except Exception:
            # not mounted yet
            pass

        container = Container(
            Static(f"Task: {new_task}"),
            Input(
                placeholder="Type your response here",
            ),
            id="respond",
        )

        # mount the container
        await self.mount(container)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        # Update the details panel with the selected item
        self.selected_task = event.button.label
        self.selected_task_id = event.button.task_id

    async def on_input_submitted(self, event: Input.Submitted) -> None:
        response = HumanResponse(result=event.value).model_dump()
        async with httpx.AsyncClient(timeout=120.0) as client:
            await client.post(
                f"{self.human_service_url}/tasks/{self.selected_task_id}/handle",
                json=response,
            )

        # remove the input container
        await self.query_one("#respond").remove()

        # remove the task from the list
        new_tasks = [
            task for task in self.tasks if task.task_id != self.selected_task_id
        ]
        self.tasks = [*new_tasks]



import httpx
import logging
import pprint
from typing import Any, Optional

from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.reactive import reactive
from textual.widgets import Button, Header, Footer, Static, Input

from llama_agents.app.components.human_list import HumanTaskList
from llama_agents.app.components.service_list import ServicesList
from llama_agents.app.components.task_list import TasksList
from llama_agents.app.components.types import ButtonType
from llama_agents.types import TaskDefinition


class LlamaAgentsMonitor(App):
    CSS = """
    Screen {
        layout: grid;
        grid-size: 2;
        grid-columns: 1fr 2fr;
        padding: 0;
    }

    #left-panel {
        width: 100%;
        height: 100%;
    }

    #right-panel {
        width: 100%;
        height: 100%;
    }

    .section {
        background: $panel;
        padding: 1;
        margin-bottom: 0;
    }

    #tasks {
        height: auto;
        max-height: 50%;
    }

    #services {
        height: auto;
        max-height: 50%;
    }

    VerticalScroll {
        height: auto;
        max-height: 100%;
        border: solid $primary;
        margin-bottom: 1;
    }

    #right-panel VerticalScroll {
        max-height: 100%;
    }

    Button {
        width: 100%;
        margin-bottom: 1;
    }

    #details {
        background: $boost;
        padding: 1;
        text-align: left;
    }

    #new-task {
        dock: bottom;
        margin-bottom: 1;
        width: 100%;
    }
    """

    details = reactive("")
    selected_service_type = reactive("")
    selected_service_url = reactive("")

    def __init__(self, control_plane_url: str, **kwargs: Any):
        self.control_plane_url = control_plane_url
        super().__init__(**kwargs)

    def compose(self) -> ComposeResult:
        yield Header()
        with Static(id="left-panel"):
            yield ServicesList(id="services", control_plane_url=self.control_plane_url)
            yield TasksList(id="tasks", control_plane_url=self.control_plane_url)
        with VerticalScroll(id="right-panel"):
            yield Static("Task or service details", id="details")
        yield Input(placeholder="Enter: New task", id="new-task")
        yield Footer()

    async def on_mount(self) -> None:
        self.set_interval(5, self.refresh_details)

    async def watch_details(self, new_details: str) -> None:
        if not new_details:
            return

        selected_type = ButtonType(new_details.split(":")[0].strip())

        if selected_type == ButtonType.SERVICE:
            self.query_one("#details").update(new_details)
        elif selected_type == ButtonType.TASK:
            self.query_one("#details").update(new_details)

    async def watch_selected_service_type(self, new_service_type: str) -> None:
        if not new_service_type:
            return

        if new_service_type == "human_service":
            await self.query_one("#right-panel").mount(
                HumanTaskList(self.selected_service_url), after=0
            )
        else:
            try:
                await self.query_one(HumanTaskList).remove()
            except Exception:
                # not mounted yet
                pass

    async def refresh_details(
        self,
        button_type: Optional[ButtonType] = None,
        selected_label: Optional[str] = None,
    ) -> None:
        if not self.details and button_type is None and selected_label is None:
            return

        selected_type = button_type or ButtonType(self.details.split(":")[0].strip())
        selected_label = (
            selected_label or self.details.split(":")[1].split("\n")[0].strip()
        )

        if selected_type == ButtonType.SERVICE:
            async with httpx.AsyncClient(timeout=120.0) as client:
                response = await client.get(
                    f"{self.control_plane_url}/services/{selected_label}"
                )
                service_def = response.json()

                service_dict = service_def
                service_url = ""
                if service_def.get("host") and service_def.get("port"):
                    service_url = f"http://{service_def['host']}:{service_def['port']}"
                    response = await client.get(f"{service_url}/")
                    service_dict = response.json()

            # format the service details nicely
            service_string = pprint.pformat(service_dict)

            self.details = (
                f"{selected_type.value}: {selected_label}\n\n{service_string}"
            )

            self.selected_service_url = service_url
            self.selected_service_type = service_dict.get("type")
        elif selected_type == ButtonType.TASK:
            async with httpx.AsyncClient(timeout=120.0) as client:
                response = await client.get(
                    f"{self.control_plane_url}/tasks/{selected_label}"
                )
                task_dict = response.json()

            # flatten the TaskResult object
            if task_dict["state"].get("result"):
                task_dict["state"]["result"] = task_dict["state"]["result"]["result"]

            # format the task details nicely
            task_string = pprint.pformat(task_dict)

            self.details = f"{selected_type.value}: {selected_label}\n\n{task_string}"
            self.selected_service_type = ""
            self.selected_service_url = ""

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        # Update the details panel with the selected item
        await self.refresh_details(
            button_type=event.button.type, selected_label=event.button.label
        )

    async def on_input_submitted(self, event: Input.Submitted) -> None:
        new_task = TaskDefinition(input=event.value).model_dump()
        async with httpx.AsyncClient(timeout=120.0) as client:
            await client.post(f"{self.control_plane_url}/tasks", json=new_task)

        # clear the input
        self.query_one("#new-task").value = ""


def run(control_plane_url: str = "http://127.0.0.1:8000") -> None:
    # remove info logging for httpx
    logging.getLogger("httpx").setLevel(logging.WARNING)

    app = LlamaAgentsMonitor(control_plane_url=control_plane_url)
    app.run()


if __name__ == "__main__":
    run()
